import boto3
import json
import logging
import sys
import os
import base64
import time
import botocore
import uuid
import re
from botocore.exceptions import ClientError
from boto3.dynamodb.conditions import Key, Attr
from datetime import datetime

# ---------------------------------------------------------------------------------
# Global Scope constants (reused across function invocations).
# ---------------------------------------------------------------------------------
region = os.environ['AWS_REGION']
ct_region = os.environ['CT_REGION']
log = logging.getLogger(__name__)
thisAcctBotoSession = boto3.session.Session()

# Create boto session to Control Tower master account using IAM role read from DB
def getCTMasterAcctSession(ctAcctRoleARN):
    thisAcctBotoSession = boto3.session.Session()
    stsClient = thisAcctBotoSession.client('sts')
    creds=stsClient.assume_role(RoleArn=ctAcctRoleARN, RoleSessionName=('RequestProcessorLambdaSession-' + str(uuid.uuid4()).replace('-',''))[:60])
    accessKey =creds['Credentials']['AccessKeyId']
    accessSecret = creds['Credentials']['SecretAccessKey']
    sessionToken = creds['Credentials']['SessionToken']
    otherAcctBotoSession = boto3.session.Session(aws_access_key_id=accessKey, aws_secret_access_key=accessSecret, aws_session_token=sessionToken, region_name=ct_region)
    return otherAcctBotoSession


# Retrieve Control Tower for a given environment.
def getCTforEnv(env):
    dynamodb = thisAcctBotoSession.resource('dynamodb', region)
    ctTable = dynamodb.Table('ControlTowers')
    dbResp = ctTable.query(IndexName="Environment-Towers", KeyConditionExpression=Key('Environment').eq(env))
    if len(dbResp['Items']) < 1:
        raise Exception('No Tower found for env: ' + env)

    # If multiple Control Towers are available for a given env then retrieve one with available capacity (< MaxAccounts environment value)
    acctTable = dynamodb.Table('CT_ACCOUNTS')
    for tower in dbResp['Items']:
        towerAcctsResp = acctTable.query(IndexName="Tower-Accounts", Select='COUNT', KeyConditionExpression=Key('TowerName').eq(tower['TowerName']))
        if towerAcctsResp['Count'] < tower['MaxAccounts']:
            if towerAcctsResp['Count'] >= int(tower['MaxAccounts'])*0.8:
                log.warning('Tower %s for environment %s is at >= 80 percent capacity.', tower['TowerName'],tower['Environment'])
            return tower

# Process Control Tower events receveid via CloudWatch EventBus.
def processCTCWEvents(event):
    log.debug(event)
    if not (event['detail']['eventSource'] == 'controltower.amazonaws.com' and event['detail']['eventName'] == 'CreateManagedAccount' and 'serviceEventDetails' in event['detail']):
        return
    
    accountName = event['detail']['serviceEventDetails']['createManagedAccountStatus']['account']['accountName']
    accountId = event['detail']['serviceEventDetails']['createManagedAccountStatus']['account']['accountId']
    accountState = event['detail']['serviceEventDetails']['createManagedAccountStatus']['state']
    
    # Update account state to reflect Service Catalog status
    if accountState == 'SUCCEEDED':
        accountState = 'AVAILABLE'
    
    dynamodb = thisAcctBotoSession.resource('dynamodb', region)
    ctAcctsTbl = dynamodb.Table('CT_ACCOUNTS')
    ctAcctsTbl.update_item(
        Key={'AccountName': accountName},
        UpdateExpression='SET AccountId = :accountId, AccountCreationState = :accountState',
        ExpressionAttributeValues={':accountId': accountId, ':accountState': accountState})

    dbResp = ctAcctsTbl.query(KeyConditionExpression=Key('AccountName').eq(accountName))
    if len(dbResp['Items']) < 1:
        log.error('Unable to find account %s in DB', accountName)
        return
    
    if accountState != 'AVAILABLE':
        log.error('Account %s state %s error.',accountName, accountState)
        return

    acctRequestId = dbResp['Items'][0]['RequestId']
    dbResp = ctAcctsTbl.query(IndexName="Request-Accounts", KeyConditionExpression=Key('RequestId').eq(acctRequestId))
    for item in dbResp['Items']:
        if not ('AccountId' in item and 'AccountCreationState' in item and item['AccountCreationState'] == 'AVAILABLE'):
            log.info('All accounts for given request Id %s not completely provisioned yet.',acctRequestId)
            return

    snsClient = thisAcctBotoSession.client('sns', region)
    snsResp = snsClient.publish(TargetArn=os.environ['SNSTopic'], 
            Message=json.dumps(dbResp['Items']), 
            Subject='Control Tower Accounts Provisioned.')

# Check to see accounts in 'UNDER_CHANGE' are now in 'AVAILABLE'/'ERROR'/'TAINTED' state. Update DB accordingly. 
def updateAccountCreationStates():
    dynamodb = thisAcctBotoSession.resource('dynamodb', region)
    ctAcctsTbl = dynamodb.Table('CT_ACCOUNTS')
    towersTbl = dynamodb.Table('ControlTowers')
    acctResp = ctAcctsTbl.query(IndexName="State-Accounts", KeyConditionExpression=Key('AccountCreationState').eq('UNDER_CHANGE'))
    for underChnageAcct in acctResp['Items']:
        towerName = underChnageAcct['TowerName']
        towersResp = towersTbl.query(KeyConditionExpression=Key('TowerName').eq(towerName))
        if len(towersResp['Items']) < 1:
            log.error('Unable to find tower: %s', towerName)
            continue
        ctAcctRoleARN = towersResp['Items'][0]['IAM_Role']
        ctMasterBotoSession = getCTMasterAcctSession(ctAcctRoleARN)
        scClient = ctMasterBotoSession.client('servicecatalog', region_name=ct_region)
        scResp = scClient.describe_provisioned_product(Id=underChnageAcct['ProvisionedProductId'])
        provProdStatus =  scResp['ProvisionedProductDetail']['Status']
        ctAcctsTbl.update_item(
            Key={'AccountName': underChnageAcct['AccountName']},
            UpdateExpression='SET AccountCreationState = :accountCreationState',
            ExpressionAttributeValues={':accountCreationState': provProdStatus})
        print(scResp)

# Use Control Tower master account boto session to invoke Service Catalog APIs to create new account.
def provisionAccountFactory(accountName):
    # Retrieve Account and associated Control Tower information from DB.
    dynamodb = thisAcctBotoSession.resource('dynamodb', region)
    ctAcctsTbl = dynamodb.Table('CT_ACCOUNTS')
    towersTbl = dynamodb.Table('ControlTowers')
    acctResp = ctAcctsTbl.query(KeyConditionExpression=Key('AccountName').eq(accountName))
    if len(acctResp['Items']) < 1:
        log.error('No account found for tower %s', accountName)
        return
    acctInfo = acctResp['Items'][0]
    towersResp = towersTbl.query(KeyConditionExpression=Key('TowerName').eq(acctInfo['TowerName']))
    if len(towersResp['Items']) < 1:
        log.error('No tower found with name %s', acctInfo['TowerName'])
        return
    towerInfo = towersResp['Items'][0]

    # Setup STS session with target Control Tower Master Account to create Service Catalog boto client.
    otherAcctBotoSession = getCTMasterAcctSession(towerInfo['IAM_Role'])
    scClient = otherAcctBotoSession.client('servicecatalog', region_name=ct_region)

    # Retrieve Control Tower Account Factory Product Id
    scResp = scClient.search_products(Filters={'FullTextSearch': ['Control Tower Account Factory']})
    productId = scResp['ProductViewSummaries'][0]['ProductId']
    scResp = scClient.list_provisioning_artifacts(ProductId=productId)

    # Retrieve 'Active' Provisioning artifact Id for a given Account Factory Service Catalog Product.
    provisioningArtifacts = [provisioningArtifact for provisioningArtifact in scResp['ProvisioningArtifactDetails'] if provisioningArtifact['Active']]
    if len(provisioningArtifacts) == 0:
        log.error('No provisioning artifacts found.')
        raise Exception("No provisioning artifacts found.")
    provisioningArtifactId = provisioningArtifacts[0]['Id']

    # Prepare ProvisioningParams to provision Service Catalog Account Factory Product.
    provisioningParams = [
        { 'Key': 'AccountName', 'Value': acctInfo['AccountName'] }, 
        { 'Key': 'AccountEmail', 'Value': acctInfo['AccountEmail'] },
        { 'Key': 'ManagedOrganizationalUnit', 'Value': acctInfo['ManagedOU'] }, 
        { 'Key': 'SSOUserFirstName', 'Value': acctInfo['SSOUserFirstName'] },
        { 'Key': 'SSOUserLastName', 'Value': acctInfo['SSOUserLastName'] }, 
        { 'Key': 'SSOUserEmail', 'Value': acctInfo['SSOUserEmail'] }
    ]
    # Initiate privisoning of Service Catalog Account Factory Product.
    scResp = scClient.provision_product(ProductId=productId, ProvisioningArtifactId=provisioningArtifactId, ProvisionedProductName=acctInfo['ProvisonedProductName'], 
        ProvisioningParameters=provisioningParams)
    if 'RecordDetail' not in scResp or 'ProvisionedProductId' not in scResp['RecordDetail']:
        log.error('Unable to retrieve ProvisionedProductId.')
        return
    
    # Update Account in DB to indicate 'UNDER_CHANGE' state so no other account provisioning is initiated until it is finished or fails.
    # Log CreatedTime and ProvisionedProductId to the Account in DB.
    ctAcctsTbl.update_item(
        Key={'AccountName': accountName},
        UpdateExpression='SET ProvisionedProductId = :provisionedProductId, AccountCreateStart = :accountCreateStart, AccountCreationState = :accountCreationState',
        ExpressionAttributeValues={
            ':provisionedProductId': scResp['RecordDetail']['ProvisionedProductId'], 
            ':accountCreateStart': str(scResp['RecordDetail']['CreatedTime']), 
            ':accountCreationState': 'UNDER_CHANGE'})
    
    log.debug("Provisioned new account in Control Tower with ProvisionedProductId={}".format(scResp['RecordDetail']['ProvisionedProductId']))

def provisionProductOneAtATime():
    # Check accounts in 'UNDER_CHANGE' to be in 'AVAILABLE'/'ERROR'/'TAINTED' state and update DB.
    updateAccountCreationStates()
    # Retrieve all 'PENDING' and 'UNDER_CHANGE' accounts.
    dynamodb = thisAcctBotoSession.resource('dynamodb', region)
    ctAcctsTbl = dynamodb.Table('CT_ACCOUNTS')
    acctList = []
    acctResp = ctAcctsTbl.query(IndexName="State-Accounts", KeyConditionExpression=Key('AccountCreationState').eq('PENDING'))
    acctList.extend(acctResp['Items'])
    acctResp = ctAcctsTbl.query(IndexName="State-Accounts", KeyConditionExpression=Key('AccountCreationState').eq('UNDER_CHANGE'))
    acctList.extend(acctResp['Items'])
    # Prepare map with list of 'PENDING' and 'UNDER_CHANGE' accounts for each Control Tower
    towerAcctMap = {}
    for acct in acctList:
        towerName = acct['TowerName']
        if towerName not in towerAcctMap:
            towerAcctMap[towerName] = {}
        acctCreationState = acct['AccountCreationState']
        if acctCreationState not in towerAcctMap[towerName]:
            towerAcctMap[towerName][acctCreationState] = []
        towerAcctMap[towerName][acctCreationState].append(acct['AccountName'])
    # Process the Map to provision 'PENDING' accounts.
    for towerName in towerAcctMap.keys():
        # Cannot initiate new Account if existing account is under 'UNDER_CHANGE' for a given Control Tower.
        if 'UNDER_CHANGE' in towerAcctMap[towerName]:
            continue
        # Make sure there is at lease one 'PENDING' account for a given Control Tower
        if 'PENDING' not in towerAcctMap[towerName]:
            continue
        # Initiate Account creation for the first-oldest 'PENDING' account.
        acctToProvision = towerAcctMap[towerName]['PENDING'][0]
        provisionAccountFactory(acctToProvision)

def lambda_handler(event, context):
    try:
        logLevel = os.environ['LogLevel']
        if not logLevel:
            log.setLevel(20)
        else:
            log.setLevel(int(logLevel))

        log.debug(event)
        log.info(json.dumps(event))

        # Handle Control Tower CloudWatch Events sent from Control Tower Master account to Control Plane account (this).
        if 'detail' in event and 'eventSource' in event['detail'] and 'eventName' in event['detail']:
            log.debug('Handle CloudWatch Events from Control Tower Master accounts.')
            processCTCWEvents(event)
            provisionProductOneAtATime()
            return

        # Handle CloudWatch Scheduled event to provision accounts from backlog and mark provisioned accounts in DB
        # Only one account (per Control Tower) can be provisioned at a time via Control Tower Service Catalog Account Factory Product
        if 'source' in event and 'detail-type' in event and event['source'] == 'aws.events' and event['detail-type'] == 'Scheduled Event':
            provisionProductOneAtATime()
            return

        # Load mapping between deployment model type and associated environments
        # Each environment is associated with a given Control Tower.
        messageBody = json.loads(event['Records'][0]['body'])

        deploymentToEnvMap = json.loads(os.environ['deploymentTypeToEnvMapping'])
        environments = deploymentToEnvMap[messageBody['deploymentType']]
        if not environments:
            log.error('Deployment model type specified in the request not found: ' + messageBody['deploymentType'])
            return
        log.debug('Found deployment environments %s for model %s' % (environments, messageBody['deploymentType']))

        # Iterate over each environment of current requests' deployment model. 
        for env in environments:
            log.debug(env)

            # Retrieve Control Tower properties associated with current environment
            towerObj = getCTforEnv(env)
            ctAcctRoleARN = towerObj['IAM_Role']
            ctAccountId = ctAcctRoleARN.split(':')[4]

            # Check to make sure account with given name and environment value does not already exists in the DB
            dynamodb = thisAcctBotoSession.resource('dynamodb', region)
            acctTbl = dynamodb.Table('CT_ACCOUNTS')
            accountName = messageBody['applicationName']+'-'+env
            dbResp = acctTbl.query(KeyConditionExpression=Key('AccountName').eq(accountName))
            if len(dbResp['Items']) > 0:
                log.error('Account with name %s already exists.', accountName)
                return
            # Validate account email address and generate unique email by adding in environment value and application name.
            emailArr = messageBody['accountEmail'].split('@')
            if len(emailArr) < 2:
                log.error('Invalid account email: %s', messageBody['accountEmail'])
                return
            accountEmail = emailArr[0] + '+' + messageBody['applicationName'] + env + '@' + emailArr[1]
            managedOU = towerObj['ManagedOU']
            ssoUserFirstName = messageBody['ssoUserFirstName']
            ssoUserLastName = messageBody['ssoUserLastName']
            ssoUserEmail = messageBody['ssoUserEmail']
            # Insert account info into the DB in 'PENDING' state.
            acctItem = {
                "AccountName": accountName,
                "RequestId": event['Records'][0]["messageId"],
                "TowerName": towerObj["TowerName"],
                "Environment": env,
                "DeploymentType": messageBody['deploymentType'],
                "Create_DateTime": datetime.now().isoformat(),
                "AccountEmail": accountEmail,
                "ManagedOU": managedOU,
                "SSOUserFirstName": ssoUserFirstName,
                "SSOUserLastName": ssoUserLastName,
                "SSOUserEmail": ssoUserEmail,
                "ApplicationName": messageBody['applicationName'],
                "ProvisonedProductName": messageBody['applicationName']+'-'+env+'-'+str(uuid.uuid4()).replace('-',''),
                "AccountCreationState": 'PENDING'
            }
            acctTbl.put_item(Item=acctItem)
            # Provision account one at a time per Control Tower.
            provisionProductOneAtATime()

    except Exception as err:
        log.error('Error in RequestProcessor lambda (%s)', err)